"""
helpers for the pytest suite
"""
import json
import sys
from colorama import init, Fore, Style


init()


def dbg(text):
    """debug printer for tests"""
    if isinstance(text, dict):
        text = json.dumps(text, sort_keys=True, indent=2)
    caller = sys._getframe(1)
    print("")
    print(Fore.GREEN + Style.BRIGHT)
    print("----- {} line {} ------".format(caller.f_code.co_name, caller.f_lineno))
    print(text)
    print("-----")
    print(Style.RESET_ALL)
    print("")



"""
text manipulation utilities
"""
import re
from datetime import datetime


DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
DATE_FORMAT_MILLIS = "%Y-%m-%dT%H:%M:%S.%fZ"
FIRST_CAP = re.compile("(.)([A-Z][a-z]+)")
ALL_CAP = re.compile("([a-z0-9])([A-Z])")
LOCALS_FILTER = ["self", "kwargs"]


def snakeify(text):
    """camelCase to snake_case"""
    first_string = FIRST_CAP.sub(r"\1_\2", text)
    return ALL_CAP.sub(r"\1_\2", first_string).lower()


def filter_locals(local_variables, extras=None):
    """filters out builtin variables in the local scope and returns locals as a dict"""
    var_filter = LOCALS_FILTER.copy()
    if extras and isinstance(extras, list):
        var_filter += extras
    return {
        x: local_variables[x]
        for x in local_variables
        if local_variables[x] is not None and x not in var_filter
    }


def datetime_to_iso8601(date_object, milliseconds=False):
    """converts a datetime to the time format bugsnag wants"""
    return date_object.strftime(DATE_FORMAT_MILLIS if milliseconds else DATE_FORMAT)


def iso8601_to_datetime(date_string, milliseconds=False):
    """converts a iso8601 string to a python datetime"""
    return datetime.strptime(
        date_string, DATE_FORMAT_MILLIS if milliseconds else DATE_FORMAT
    )


def dict_to_query_params(params):
    """given a dictionary of query params, form the query param string"""
    if not params:
        return ""
    for param in params:
        if isinstance(params[param], bool):
            params[param] = str(params[param]).lower()
        elif isinstance(params[param], datetime):
            params[param] = datetime_to_iso8601(params[param])
    return "?{}".format("&".join(["{}={}".format(x, params[x]) for x in params]))



"""
models for each object in the bugsnag data access api
"""
import json
from datetime import datetime
from pybugsnag.globals import LIBRARY
from pybugsnag.utils.text import (
    filter_locals,
    snakeify,
    dict_to_query_params,
    datetime_to_iso8601,
    iso8601_to_datetime,
)


class BaseModel:
    """basic model that just parses camelCase json to snake_case keys"""

    def __init__(self, data, client=None, **kwargs):
        """constructor"""
        self._data = {**data, **kwargs}
        self._json = self._jsond(data)
        self._client = client

        for key in self._data:
            setattr(self, snakeify(key), self._data[key])

    def _jsond(self, json_data):
        """json dumps"""
        return json.dumps(json_data)

    def _jsonl(self, dictionary):
        """json loads"""
        return json.loads(dictionary)


class Collaborator(BaseModel):
    """bugsnag user (collaborator) object"""

    def __init__(self, data, **kwargs):
        """override"""
        super(Collaborator, self).__init__(data, **kwargs)

    def __repr__(self):
        """repr"""
        return "<{}.Collaborator[{}] '{}'>".format(LIBRARY, self.id, self.email)


class Event(BaseModel):
    """bugsnag event object"""

    DATE_FIELDS = ["received_at"]

    class Sort:
        """event sort enum"""

        class Direction:
            """event sort direction enum"""

            ASCENDING = "asc"
            DESCENDING = "desc"

        TIMESTAMP = "timestamp"

    def __init__(self, data, **kwargs):
        """override"""
        super(Event, self).__init__(data, **kwargs)
        for field in Event.DATE_FIELDS:
            setattr(
                self,
                field,
                iso8601_to_datetime(getattr(self, field), milliseconds=True),
            )

    def __repr__(self):
        """repr"""
        return "<{}.Event[{}] '{}'>".format(LIBRARY, self.id, self.context)


class EventField(BaseModel):
    """bugsnag eventField object"""

    def __init__(self, data, **kwargs):
        """override"""
        super(EventField, self).__init__(data, **kwargs)

    def __repr__(self):
        """repr"""
        return "<{}.EventField[{}] '{}'>".format(
            LIBRARY, self.display_id, self.filter_options["name"]
        )


class Pivot(BaseModel):
    """bugsnag pivot object"""

    def __init__(self, data, **kwargs):
        """override"""
        super(Pivot, self).__init__(data, **kwargs)

    def __repr__(self):
        """repr"""
        return "<{}.Pivot[{}] '{}'>".format(
            LIBRARY, self.event_field_display_id, self.name
        )


class Error(BaseModel):
    """bugsnag error object"""

    MIN_BUCKETS = 1
    MAX_BUCKETS = 50
    DATE_FIELDS = ["first_seen", "last_seen", "first_seen_unfiltered"]

    class Resolution:
        """time resolution enum"""

        ONE_MINUTE = "1m"
        FIVE_MINUTE = "5m"
        THIRTY_MINUTE = "30m"
        TWO_HOUR = "2h"
        TWELVE_HOUR = "12h"

    class Sort:
        """error sort enum"""

        class Direction:
            """error sort direction enum"""

            ASCENDING = "asc"
            DESCENDING = "desc"

        LAST_SEEN = "last_seen"
        FIRST_SEEN = "first_seen"
        USERS = "users"
        EVENTS = "events"
        UNSORTED = "unsorted"

    class Severity:
        """severity enum"""

        ERROR = "error"
        WARNING = "warning"
        INFO = "info"

    def __init__(self, data, **kwargs):
        """override"""
        super(Error, self).__init__(data, **kwargs)
        for field in Error.DATE_FIELDS:
            setattr(self, field, iso8601_to_datetime(getattr(self, field)))

    def __repr__(self):
        """repr"""
        return "<{}.Error[{}] '{}'>".format(LIBRARY, self.id, self.error_class)

    def get_event(self, event_id):
        """gets an event by id for this error"""
        return Event(
            self._client.get("projects/{}/events/{}".format(self.project.id, event_id)),
            client=self._client,
            project=self.project,
            error=self,
        )

    def get_latest_event(self):
        """gets the latest event for this error"""
        return Event(
            self._client.get("errors/{}/latest_event".format(self.id)),
            client=self._client,
            project=self.project,
            error=self,
        )

    def get_events(
        self,
        base=None,
        sort=Event.Sort.TIMESTAMP,
        direction=Event.Sort.Direction.DESCENDING,
        per_page=30,
        filters=None,  # TODO
        full_reports=False,
        **kwargs
    ):
        """get events for this error"""
        params = filter_locals(locals())

        if "base" not in params:
            params["base"] = datetime.now()

        query_params = dict_to_query_params(params)
        path = "projects/{}/errors/{}/events{}".format(
            self.project.id, self.id, query_params
        )
        return [
            Event(x, error=self, client=self._client, project=self.project)
            for x in self._client.get(path)
        ]

    def get_trend_buckets(self, buckets_count=10):
        """get trend buckets for this error"""
        buckets = min(max(Error.MIN_BUCKETS, buckets_count), Error.MAX_BUCKETS)
        return self._client.get(
            "projects/{}/errors/{}/trend?buckets_count={}".format(
                self.project.id, self.id, buckets
            )
        )

    def get_trend_resolution(self, resolution=None):
        """get trend buckets for this error based on time resolution"""
        if not resolution:
            resolution = Error.Resolution.FIVE_MINUTE
        return self._client.get(
            "projects/{}/errors/{}/trend?resolution={}".format(
                self.project.id, self.id, resolution
            )
        )

    def get_pivots(self, summary_size=10, per_page=30):
        """get a number of pivots for this error"""
        params = filter_locals(locals())
        query_params = dict_to_query_params(params)
        return [
            Pivot(x, project=self.project, error=self, client=self._client)
            for x in self._client.get(
                "projects/{}/errors/{}/pivots{}".format(
                    self.project.id, self.id, query_params
                )
            )
        ]


class Release(BaseModel):
    """bugsnag release object"""

    class Sort:
        """release sort enum"""

        class Direction:
            """release sort direction enum"""

            ASCENDING = "asc"
            DESCENDING = "desc"

        TIMESTAMP = "timestamp"
        PERCENT_OF_SESSIONS = "percent_of_sessions"
        CRASH_RATE = "crash_rate"
        PERCENT_OF_EVENTS = "percent_of_events"

    def __init__(self, data, **kwargs):
        """override"""
        super(Release, self).__init__(data, **kwargs)

    def __repr__(self):
        """repr"""
        return "<{}.Release[{}] '{}'>".format(LIBRARY, self.id, self.app_version)


class Project(BaseModel):
    """bugsnag project object"""

    MIN_BUCKETS = 1
    MAX_BUCKETS = 50
    DATE_FIELDS = ["created_at", "updated_at"]

    class Resolution:
        """time resolution enum"""

        ONE_MINUTE = "1m"
        FIVE_MINUTE = "5m"
        THIRTY_MINUTE = "30m"
        TWO_HOUR = "2h"
        TWELVE_HOUR = "12h"

    class Sort:
        """project sort enum"""

        class Direction:
            """project sort direction enum"""

            ASCENDING = "asc"
            DESCENDING = "desc"

        CREATED_AT = "created_at"

    def __init__(self, data, **kwargs):
        """override"""
        super(Project, self).__init__(data, **kwargs)
        for field in Project.DATE_FIELDS:
            setattr(
                self,
                field,
                iso8601_to_datetime(getattr(self, field), milliseconds=True),
            )

    def __repr__(self):
        """repr"""
        return "<{}.Project[{}] '{}'>".format(LIBRARY, self.id, self.name)

    def get_error(self, error_id):
        """gets an error by id for this project"""
        return Error(
            self._client.get("projects/{}/errors/{}".format(self.id, error_id)),
            client=self._client,
            project=self,
        )

    def get_errors(
        self,
        base=None,
        sort=Error.Sort.LAST_SEEN,
        direction=Error.Sort.Direction.DESCENDING,
        per_page=30,
        filters=None,  # TODO
        **kwargs
    ):
        """get errors for this project"""
        params = filter_locals(locals())

        if "base" not in params:
            params["base"] = datetime.now()

        query_params = dict_to_query_params(params)
        path = "projects/{}/errors{}".format(self.id, query_params)
        return [
            Error(x, project=self, client=self._client) for x in self._client.get(path)
        ]

    def get_event(self, event_id):
        """gets an event by id for this project"""
        return Event(
            self._client.get("projects/{}/events/{}".format(self.id, event_id)),
            client=self._client,
            project=self,
        )

    def get_events(
        self,
        base=None,
        sort=Event.Sort.TIMESTAMP,
        direction=Event.Sort.Direction.DESCENDING,
        per_page=30,
        filters=None,  # TODO
        full_reports=False,
        **kwargs
    ):
        """get events for this project"""
        params = filter_locals(locals())

        if "base" not in params:
            params["base"] = datetime.now()

        query_params = dict_to_query_params(params)
        path = "projects/{}/events{}".format(self.id, query_params)
        return [
            Event(x, project=self, client=self._client) for x in self._client.get(path)
        ]

    def get_trend_buckets(self, buckets_count=10):
        """get trend buckets for this project"""
        buckets = min(max(Project.MIN_BUCKETS, buckets_count), Project.MAX_BUCKETS)
        return self._client.get(
            "projects/{}/trend?buckets_count={}".format(self.id, buckets)
        )

    def get_trend_resolution(self, resolution=None):
        """get trend buckets for this project based on time resolution"""
        if not resolution:
            resolution = Project.Resolution.FIVE_MINUTE
        return self._client.get(
            "projects/{}/trend?resolution={}".format(self.id, resolution)
        )

    def get_release(self, release_id):
        """get a single release by id"""
        return Release(
            self._client.get("projects/{}/releases/{}".format(self.id, release_id)),
            project=self,
            client=self._client,
        )

    def get_releases(
        self,
        release_stage=None,
        base=None,
        sort=Release.Sort.TIMESTAMP,
        offset=0,
        per_page=5,
    ):
        """get a list of releases for this project"""
        params = filter_locals(locals())

        if "base" in params:
            params["base"] = datetime_to_iso8601(params["base"])

        query_params = dict_to_query_params(params)
        path = "projects/{}/releases{}".format(self.id, query_params)
        return [
            Release(x, project=self, client=self._client)
            for x in self._client.get(path)
        ]

    def get_pivots(self, summary_size=10):
        """get a number of pivots for this project"""
        params = filter_locals(locals())
        query_params = dict_to_query_params(params)
        return [
            Pivot(x, project=self, client=self._client)
            for x in self._client.get(
                "projects/{}/pivots{}".format(self.id, query_params)
            )
        ]

    def get_event_fields(self):
        """gets a list of event fields for the project"""
        return [
            EventField(x, project=self, client=self._client)
            for x in self._client.get("projects/{}/event_fields".format(self.id))
        ]


class Organization(BaseModel):
    """bugsnag organization object"""

    DATE_FIELDS = ["created_at", "updated_at"]

    def __init__(self, data, **kwargs):
        """override"""
        super(Organization, self).__init__(data, **kwargs)
        self._projects = None
        self._collaborators = None
        self._admins_count = None
        for field in Organization.DATE_FIELDS:
            setattr(
                self,
                field,
                iso8601_to_datetime(getattr(self, field), milliseconds=True),
            )

    def __repr__(self):
        """repr"""
        return "<{}.Organization[{}] '{}'>".format(LIBRARY, self.id, self.name)

    @property
    def projects(self):
        """cachable projects property"""
        if not self._projects or not self._client.cache:
            self._projects = self.get_projects()
        return self._projects

    @property
    def collaborators(self):
        """cachable collaborators property"""
        if not self._collaborators or not self._client.cache:
            self._collaborators = self.get_collaborators()
        return self._collaborators

    @property
    def admins_count(self):
        """gets the count of admin collaborators"""
        if not self._admins_count or not self._client.cache:
            self._admins_count = int(
                self._client.get(
                    "organizations/{}/admins_count".format(self.id), raw=True
                ).text
            )
        return self._admins_count

    def get_projects(
        self,
        sort=Project.Sort.CREATED_AT,
        direction=Project.Sort.Direction.DESCENDING,
        per_page=30,
    ):
        """gets the projects based on the params"""
        params = filter_locals(locals())
        query_params = dict_to_query_params(params)

        path = "organizations/{}/projects{}".format(self.id, query_params)
        return [
            Project(x, organization=self, client=self._client)
            for x in self._client.get(path)
        ]

    def get_collaborators(self, per_page=30, q=None, exclude_project=None):
        """get collaborators for this organization"""
        params = filter_locals(locals())
        query_params = dict_to_query_params(params)

        path = "organizations/{}/collaborators{}".format(self.id, query_params)
        return [
            Collaborator(x, organization=self, client=self._client)
            for x in self._client.get(path)
        ]

    def get_collaborator(self, collaborator_id):
        """get collaborator by id"""
        return Collaborator(
            self._client.get(
                "organizations/{}/collaborators/{}".format(self.id, collaborator_id)
            ),
            organization=self,
            client=self._client,
        )



"""
globals
"""


__version__ = "0.0.5"


LIBRARY = "pybugsnag"
API_URL = "https://api.bugsnag.com/"


TEST_API_URL = "https://private-anon-3633b611b0-bugsnagapiv2.apiary-mock.com/"
TEST_TOKEN = "access_token"



"""
error models for pybugsnag
"""


class PyBugsnagException(Exception):
    """base pybugsnag exception class"""

    def __init__(self, *args, **kwargs):
        extra = ""
        if args:
            extra = '\n| extra info: "{extra}"'.format(extra=args[0])
        print(
            "[{exception}]: {doc}{extra}".format(
                exception=self.__class__.__name__, doc=self.__doc__, extra=extra
            )
        )
        Exception.__init__(self, *args, **kwargs)


class RateLimited(PyBugsnagException):
    """request received a 429 - you are currently rate limited"""



"""
configure pytest
"""
import pytest
from pybugsnag.test.helpers import dbg


@pytest.fixture(scope="session", autouse=True)
def before_all(request):
    """test setup"""
    dbg("[+] begin pybugsnag tests")
    request.addfinalizer(after_all)


def after_all():
    """tear down"""
    dbg("[+] end pybugsnag tests")



"""
a base test suite for pybugsnag
"""
from datetime import datetime
from pybugsnag.models import (
    Collaborator,
    Error,
    Event,
    EventField,
    Organization,
    Pivot,
    Project,
    Release,
)
from pybugsnag.models.client import test_client


CLIENT = test_client()


def is_list_of_type(check_list, check_type):
    """helper function for checking if it's a list and of a specific type"""
    assert isinstance(check_list, list)
    assert isinstance(check_list[0], check_type)
    return True


def test_organizations():
    """testing accessing organizations"""
    organizations = CLIENT.organizations
    assert organizations
    assert isinstance(organizations, list)

    organization = organizations[0]
    assert organization
    assert isinstance(organization, Organization)

    found_organization = CLIENT.get_organization(organization.id)
    assert isinstance(found_organization, Organization)

    assert isinstance(organization.created_at, datetime)
    assert isinstance(organization.updated_at, datetime)

    assert organization.admins_count == 5
    assert is_list_of_type(organization.collaborators, Collaborator)
    assert is_list_of_type(organization.projects, Project)

    found_collaborator = organization.get_collaborator(organization.collaborators[0].id)
    assert isinstance(found_collaborator, Collaborator)

    assert "<pybugsnag.Collaborator" in str(found_collaborator)
    assert "<pybugsnag.Organization" in str(organization)


def test_projects():
    """testing features around projects"""
    organization = CLIENT.organizations[0]
    assert organization
    assert is_list_of_type(organization.projects, Project)
    project = organization.projects[0]

    found_project = CLIENT.get_project(project.id)
    assert isinstance(found_project, Project)

    assert isinstance(project.created_at, datetime)
    assert isinstance(project.updated_at, datetime)

    errors = project.get_errors()
    assert is_list_of_type(errors, Error)

    error_found = project.get_error(errors[0].id)
    assert isinstance(error_found, Error)

    events = project.get_events()
    assert is_list_of_type(events, Event)

    event_found = project.get_event(events[0].id)
    assert isinstance(event_found, Event)

    releases = project.get_releases()
    assert is_list_of_type(releases, Release)

    # TODO reimplement when the apiary is fixed
    # release_found = project.get_release(releases[0].id)
    # assert isinstance(release_found, Release)

    trend_buckets = project.get_trend_buckets()
    assert is_list_of_type(trend_buckets, dict)

    trend_resolution = project.get_trend_resolution()
    assert is_list_of_type(trend_resolution, dict)

    pivots = project.get_pivots()
    assert is_list_of_type(pivots, Pivot)

    event_fields = project.get_event_fields()
    assert is_list_of_type(event_fields, EventField)

    assert "<pybugsnag.Release" in str(releases[0])
    assert "<pybugsnag.EventField" in str(event_fields[0])


def test_errors():
    """testing features around errors"""
    organization = CLIENT.organizations[0]
    assert organization
    assert isinstance(organization.projects, list)

    project = organization.projects[0]
    assert isinstance(project, Project)

    errors = project.get_errors()
    assert is_list_of_type(errors, Error)

    error = errors[0]
    events = error.get_events()
    assert is_list_of_type(events, Event)
    assert isinstance(events[0].project, Project)
    assert isinstance(error.first_seen, datetime)
    assert isinstance(error.last_seen, datetime)
    assert isinstance(error.first_seen_unfiltered, datetime)

    latest_event = error.get_latest_event()
    assert isinstance(latest_event, Event)

    found_event = error.get_event(latest_event.id)
    assert isinstance(found_event, Event)

    trend_buckets = error.get_trend_buckets()
    assert is_list_of_type(trend_buckets, dict)

    trend_resolution = error.get_trend_resolution()
    assert is_list_of_type(trend_resolution, dict)

    pivots = error.get_pivots()
    assert is_list_of_type(pivots, Pivot)

    assert "<pybugsnag.Error" in str(error)
    assert "<pybugsnag.Pivot" in str(pivots[0])


def test_events():
    """testing features around events"""
    organization = CLIENT.organizations[0]
    assert organization
    assert isinstance(organization.projects, list)

    project = organization.projects[0]
    assert isinstance(project, Project)

    errors = project.get_errors()
    assert is_list_of_type(errors, Error)
    error = errors[0]

    events = error.get_events()
    assert is_list_of_type(events, Event)
    event = events[0]

    assert isinstance(event.received_at, datetime)
    assert "<pybugsnag.Event" in str(event)



"""
base client model to create and use http endpoints
"""
import requests
import urllib.parse
from pybugsnag.globals import __version__, API_URL, LIBRARY, TEST_TOKEN, TEST_API_URL
from pybugsnag.models.error import RateLimited
from pybugsnag.models import Organization, Project


def test_client():
    """returns a test client"""
    return BugsnagDataClient(TEST_TOKEN, api_url=TEST_API_URL, debug=True)


class BugsnagDataClient:
    """client http wrapper"""

    def __init__(self, token, api_url=API_URL, cache=True, debug=False):
        """creates a new client"""
        if not token:
            raise Exception("no token specified!")
        self.token = token
        self.api_url = api_url
        self.version = __version__
        self.cache = cache
        self.debug = debug

        # cache
        self._organizations = None

    @property
    def headers(self):
        """forms the headers required for the API calls"""
        return {
            "Accept": "application/json; version=2",
            "AcceptEncoding": "gzip, deflate",
            "Authorization": "token {}".format(self.token),
            "User-Agent": "{}/{}".format(LIBRARY, self.version),
        }

    def _log(self, *args):
        """logging method"""
        if not self.debug:
            return
        print(*args)

    def _req(self, path, method="get", **kwargs):
        """requests wrapper"""
        full_path = urllib.parse.urljoin(self.api_url, path)
        self._log("[{}]: {}".format(method.upper(), full_path))
        request = requests.request(method, full_path, headers=self.headers, **kwargs)
        if request.status_code == 429:
            raise RateLimited()
        return request

    def get(self, path, raw=False, **kwargs):
        """makes a get request to the API"""
        request = self._req(path, **kwargs)
        return request if raw else request.json()

    def post(self, path, raw=False, **kwargs):
        """makes a post request to the API"""
        request = self._req(path, method="post", **kwargs)
        return request if raw else request.json()

    def put(self, path, raw=False, **kwargs):
        """makes a put request to the API"""
        request = self._req(path, method="put", **kwargs)
        return request if raw else request.json()

    @property
    def organizations(self):
        """organizations list for this access token"""
        if not self._organizations or not self.cache:
            self._organizations = [
                Organization(x, client=self) for x in self.get("user/organizations")
            ]
        return self._organizations

    def get_organization(self, organization_id):
        """get organization info by organization_id"""
        return Organization(
            self.get("organizations/{}".format(organization_id)), client=self
        )

    def get_project(self, project_id):
        """gets a project by it's id"""
        return Project(self.get("projects/{}".format(project_id)), client=self)



#!/usr/bin/env python
"""
pip setup file
"""
from setuptools import setup, find_packages

LIBRARY = "pybugsnag"

with open("README.rst") as readme:
    long_description = readme.read()


with open("requirements.txt") as requirements:
    install_requires = requirements.read().split("\n")
    install_requires = [x.strip() for x in install_requires if x.strip()]

def find_version(*file_paths):
    import os
    import re
    """
    This pattern was modeled on a method from the Python Packaging User Guide:
        https://packaging.python.org/en/latest/single_source_version.html
    We read instead of importing so we don't get import errors if our code
    imports from dependencies listed in install_requires.
    """
    base_module_file = os.path.join(*file_paths)
    with open(base_module_file) as f:
        base_module_data = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              base_module_data, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name=LIBRARY,
    version=find_version('pybugsnag', 'globals.py'),
    description="A python wrapper for the Bugsnag Data Access API",
    long_description=long_description,
    author="Jacobi Petrucciani",
    author_email="jacobi@mimirhq.com",
    keywords="bugsnag python data api",
    url="https://github.com/jpetrucciani/{}.git".format(LIBRARY),
    download_url="https://github.com/jpetrucciani/{}.git".format(LIBRARY),
    license="LICENSE",
    packages=find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    zip_safe=False,
)
